//@line 36 "/builds/tinderbox/Fx-Trunk-l10n-Release/Linux_2.6.18-8.el5_Depend/l10n/ka/browser/firefox-l10n.js"

//@line 38 "/builds/tinderbox/Fx-Trunk-l10n-Release/Linux_2.6.18-8.el5_Depend/l10n/ka/browser/firefox-l10n.js"

pref("general.useragent.locale", "ka");
